/*
 *  seed.cc created on Mon, 31 May 1999 21:12:51 -0400 by noe
 */    
#ifndef __SEED_H__
#define __SEED_H__

//STL
//#include <vector>
//#include <list>
//STD
#include <iostream>
#include <fstream>
#include <cstdio>
#include <cstdlib>
#include <cmath>
#include <string>
//STR
#include "macro.h"
using namespace std;
         
class seed {
 public:
  
  seed(){};
  //seed(string & m){};
  seed(int weight,int span,int nbTrans=0, bool noBorderJokers=false):_weight(weight),_span(span),_nbTrans(nbTrans),_noBorderJokers(noBorderJokers){
    
    // build a first seed without transitions of span "span - nbTrans"
    _seedMotifWithoutTrans = string(weight,'#');
    _seedMotifWithoutTrans.append(string(span-nbTrans-weight,'-'));
    
    // build a second seed from the first one with transition elements at the end
    _seedMotif = string(_seedMotifWithoutTrans);
    _seedMotif.append(string(nbTrans,'@'));         
  }
  
  ~seed(){};

  /** @name Output functions
   *  @brief these methods are provided to get current seed processed
   */
  
  //@{

  /** @brief gives the current seed representation (a word in  {#,-,@})
   *  @return a char * table (old string representation)
   */
  const char *   c_str() {return _seedMotif.c_str();};
  
  /** @brief gives the current seed representation (a word in  {#,-,@})
   *  @return a string object
   */
  string &         str(){return _seedMotif;};

  
  //@}


  /** @brief generates the next seed 
   *  @return 1 if next seed has been generated, 0 if all the seeds in the set have been generated
   */
  int next(bool degenerated = false);

  /** @brief generates a random seed 
   *  @return 1
   */
  int randomized(bool degenerated = false);
  
  
  /** @brief return the 2-profile of the seed
   *  @return nothing but you must give correct parameters to 
   */
  int profile(/* out */int * profile_size, /* out */ int ** profile);
  
  bool profile_eq( /* out */seed& other)
    {
      int result;
      int as,*a,bs,*b;
      this->profile(&as,&a);
      other.profile(&bs,&b);
      if (as == bs){
	for(int i=0;i<as;i++){
	  if (a[i] !=  b[i]){
	    result = 0;
	    goto en;
	  }
	}
	result = 1;
      }else{
	result = 0;
      }
    en:
      delete[] a;
      delete[] b;
      return result;
    }
  
  bool profile_geq( /* out */seed& other)
    {
      int result = 0,delta = 0;
      int as,*a,bs,*b;
      this->profile(&as,&a);
      other.profile(&bs,&b);
      if (as == bs){
	for(int i=0;i<as;i++){
	  delta += a[i] - b[i];
	  result += delta;
	}
      }else{
	result = 0;
      }
      delete[] a;
      delete[] b;
      return result;
    }
  

 protected:
  /// Current seed motif represented as a word in {#,-,@}
  string _seedMotif;
  /// Seed motif represented as a word in {#,-}. Corresponds to _seedMotif without its "@" elements	
  string _seedMotifWithoutTrans;
  /// Fixed seed "#"-weight 
  int    _weight;
  /// Fixed seed span 
  int    _span;
  /// Fixed Number of transitions
  int    _nbTrans;
  /// Do not consider seed with jokers "-" on the border as real ones ... function next() only produces non degenerated seeds
  bool   _noBorderJokers; 
  // Inner function to iterate all seeds (modify both seedMotifWithoutTrans and _seedMotif)
  inline int _next();
  /// Inner function to iterate all seeds (modify seedMotifWithoutTrans seed)
  inline int _nextWithoutTrans();  
};
#endif

